import math
import time
from threading import Thread

import numpy  as np
import pandas as pd

from aperturedb import Status
from aperturedb import ParallelLoader

ENTITY_CLASS      = "EntityClass"
CONTRAINTS_PREFIX = "constraint_"
PROPERTIES  = "properties"
CONSTRAINTS = "constraints"

class CSVParser():

    '''
        ApertureDB General CSV Parser for Loaders.
        ...
    '''

    def __init__(self, filename):

        self.df = pd.read_csv(filename)

        self.validate()

        if len(self.df) == 0:
            print("Error: Dataframe empty. Is the CSV file ok?")

        self.df = self.df.astype('object')

        self.header = list(self.df.columns.values)

    def __len__(self):

        return len(self.df.index)

    def parse_properties(self, df, idx):

        properties = {}
        if len(self.props_keys) > 0:
            for key in self.props_keys:
                # Handle Date data type
                if key.startswith("date:"):
                    prop = key[len("date:"):] # remove prefix
                    properties[prop] = {"_date": self.df.loc[idx, key]}
                else:
                    properties[key] = self.df.loc[idx, key]

        return properties

    def parse_constraints(self, df, idx):

        constraints = {}
        if len(self.constraints_keys) > 0:
            for key in self.constraints_keys:
                if key.startswith("constraint_date:"):
                    prop = key[len("constraint_date:"):] # remove prefix
                    constraints[prop] = ["==", {"_date": self.df.loc[idx, key]} ]
                else:
                    prop = key[len("constraint_"):] # remove "prefix
                    constraints[prop] = ["==", self.df.loc[idx, key]]

        return constraints

    def __getitem__(self, idx):

        Exception("__getitem__ not implemented!")

    def validate(self):

        Exception("Validation not implemented!")



import argparse
import time
import unittest

import dbinfo

from aperturedb import Connector, Status
from aperturedb import Images
from aperturedb import PyTorchDataset

class TestTorch(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # ApertureDB Server Info
        self.db_host = dbinfo.DB_HOST
        self.db_port = dbinfo.DB_PORT

        db_up = False
        attempts = 0
        while(not db_up):
            try:
                db = Connector.Connector(self.db_host, self.db_port)
                db_up = True
                if (attempts > 0):
                    print("Connection to ApertureDB successful.")
            except:
                print("Attempt", attempts,
                      "to connect to ApertureDB failed, retying...")
                attempts += 1
                time.sleep(1) # sleeps 1 second

            if attempts > 10:
                print("Failed to connect to ApertureDB after 10 attempts")
                exit()

class TestTorchDatasets(TestTorch):

    '''
        These tests need to be run after the Loaders, because it uses
        data inserted by the loaders.
    '''

    def test_omConstraints(self):

        db = Connector.Connector(self.db_host, self.db_port)

        const = Images.Constraints()
        const.greaterequal("age", 0)
        dataset = PyTorchDataset.ApertureDBDatasetConstraints(db, constraints=const)

        dbstatus = Status.Status(db)
        self.assertEqual(len(dataset), dbstatus.count_images())

        start = time.time()

        # Iterate over dataset.
        for img in dataset:
            if len(img[0]) < 0:
                print("Empty image?")
                self.assertEqual(True, False)

        print("\n")
        print("Throughput (imgs/s):", len(dataset) / (time.time() - start))

    def test_nativeContraints(self):

        db = Connector.Connector(self.db_host, self.db_port)

        query = [ {
            "FindImage": {
                "constraints": {
                    "age": [">=", 0]
                },
                "operations": [
                    {
                        "type": "resize",
                        "width": 224,
                        "height": 224
                    }
                ],
                "results": {
                    "list": ["license"]
                }
            }
        }]

        dataset = PyTorchDataset.ApertureDBDataset(db, query, label_prop="license")

        dbstatus = Status.Status(db)
        self.assertEqual(len(dataset), dbstatus.count_images())

        start = time.time()

        # Iterate over dataset.
        for img in dataset:
            if len(img[0]) < 0:
                print("Empty image?")
                self.assertEqual(True, False)

        print("\n")
        print("Throughput (imgs/s):", len(dataset) / (time.time() - start))



import os
import sys
import time

class Status(object):

    def __init__(self, connector):

        self.connector = connector

    def __repr__(self):

        return self.status()

    def status(self):

        q = [{ "GetStatus": {} }]

        try:
            res, blobs = self.connector.query(q)
        except:
            print(self.connector.get_last_response_str())

        return self.connector.get_last_response_str()

    def get_schema(self, refresh=False):

        query = [ {
            "GetSchema": {
                "type" : "entities",
                "refresh": refresh,
            }
        }]

        res, blobs = self.connector.query(query)

        schema = {}

        try:
            schema = res[0]["GetSchema"]
        except:
            print("Cannot retrieve schema")
            print(self.connector.get_last_response_str())

        return schema

    def count_images(self, constraints={}):

        q = [{
            "FindImage": {
                "blobs": False,
                "results": {
                    "count": True,
                }
            }
        }]

        if constraints:
            q[0]["FindImage"]["constraints"] = constraints

        try:
            res, blobs = self.connector.query(q)
            total_images = res[0]["FindImage"]["count"]
        except:
            total_images = 0
            print(self.connector.get_last_response_str())

        return total_images

    def count_bboxes(self, constraints={}):

        q = [{
            "FindBoundingBox": {
                "blobs": False,
                "results": {
                    "count": True,
                }
            }
        }]

        if constraints:
            q[0]["FindBoundingBox"]["constraints"] = constraints

        try:
            res, blobs = self.connector.query(q)
            total_connections = res[0]["FindBoundingBox"]["count"]
        except:
            total_connections = 0
            print(self.connector.get_last_response_str())

        return total_connections

    def count_entities(self, entity_class, constraints={}):

        q = [{
            "FindEntity": {
                "with_class": entity_class,
                "results": {
                    "count": True,
                }
            }
        }]

        if constraints:
            q[0]["FindEntity"]["constraints"] = constraints

        try:
            res, blobs = self.connector.query(q)
            total_entities = res[0]["FindEntity"]["count"]
        except:
            total_entities = 0
            print(self.connector.get_last_response_str())

        return total_entities

    def count_connections(self, connections_class, constraints={}):

        q = [{
            "FindConnection": {
                "with_class": connections_class,
                "results": {
                    "count": True,
                }
            }
        }]

        if constraints:
            q[0]["FindConnection"]["constraints"] = constraints

        try:
            res, blobs = self.connector.query(q)
            total_connections = res[0]["FindConnection"]["count"]
        except:
            total_connections = 0
            print(self.connector.get_last_response_str())

        return total_connections
